import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A thread implementation of a LockClient. The LockClient will connect
 * the servers' server sockets and start listening on its own server socket.
 * When a ClientController connects its socket, it will start passing the 
 * ClientController's commands to the servers. It will output the response
 * from the servers. LockClient is started by a StartLockClient, which will
 * its argument array to LockClient - StartLockClients' usage is:
 * java StartLockClient txnId noXLockRequests noSLockRequests ownPort
 *    server1 port1 ... serverN portN
 * The Client will ask first noSLockRequests of parameter servers for an 
 * slock and last noXLockServers for an xlock.
 * 
 * @author  Jyrki Nummenmaa
 * @version 0.1
 * @see     StartLockClient
 * @see     LockManagerProtocol
 */


public class LockClient extends Thread {

    int portNo;
    /** Server socket port for listening */
    int ownPortNo;
    /** txn id */
    long txnId = Long.MAX_VALUE;
    /** No of servers to consult for an xlock */
    int noXLockServers; 
    /** No of servers to consult for an slock */
    int noSLockServers; 
    /** No of server connections */
    int noServers;
    /** Sockets for server connections */
    Socket[] lcSockets;
    /** PrintWriters for output to servers */
    PrintWriter[] lcOuts;
    /** BufferedReaders for inputs */
    BufferedReader[] lcIns;
    /** Argument vector */
    String[] args;
    /** For storing the active lock requests (granted or queued) */
    Vector lockRequests;
    /** LockManager server host names */
    String [] serverNames;
    /** LockManager server ports */
    int [] serverPorts;

    public LockClient(String[] args) {
	this.args = args;
    }

    public void run() {
	try {

	    File serverFile = new File("servers.txt"); 
	    if (!serverFile.exists()) {
		System.out.println("Could not find servers.txt");
		System.exit(-1);
	    }
	    System.out.println("Found servers.txt");
	    FileInputStream fis = new FileInputStream(serverFile);
	    InputStreamReader isr = new InputStreamReader(fis);
	    BufferedReader br = new BufferedReader(isr);
	    String noServersString = br.readLine();
	    noServers = Integer.parseInt(noServersString);
	    serverNames = new String[noServers];
	    serverPorts = new int[noServers];

	    for (int i=0;i<noServers;i++) {
		String line = br.readLine();
		StringTokenizer st = new StringTokenizer(line);
		serverNames[i] = st.nextToken();
		serverPorts[i] = Integer.parseInt(st.nextToken());
	    }

	    System.out.println("Reading servers.txt completed");
	    
	    lcSockets = new Socket[noServers];
	    lcOuts = new PrintWriter[noServers];
	    lcIns = new BufferedReader[noServers];
	    
	    noSLockServers = Integer.parseInt(args[0]);
	    noXLockServers = noServers - noSLockServers + 1;

	    ownPortNo = Integer.parseInt(args[1]);
	    
	    lockRequests = new Vector(); // actice lock request type on item
	    	    
	    for (int i=0; i<noServers; i++) {
		lcSockets[i] = new Socket(serverNames[i],serverPorts[i]);
		lcOuts[i] = new PrintWriter(lcSockets[i].getOutputStream(), true);
		lcIns[i] = new BufferedReader(new InputStreamReader(lcSockets[i].getInputStream()));
	    }

	    // lcOuts[0].println("newtxn");
	    //for (int i=1;i<noServers;i++) {
	    //	lcOuts[i].println("txn " + txnId);
	    //}
	    
	} catch (UnknownHostException e) {
	    System.err.println("Don't know about host.");
	    System.exit(1);
	} catch (IOException e) {
	    System.err.println("Couldn't get I/O for the connection to the host.");
	    System.exit(1);
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	}


	
        ServerSocket serverSocket = null;
        try {
            serverSocket = new ServerSocket(ownPortNo);
        } catch (IOException e) {
            System.err.println("Could not listen on port.");
            System.exit(1);
        }

	// Now we have set up the connections...wait for controller


        Socket clientSocket = null;
	BufferedReader controllerIn = null;
        try {
            clientSocket = serverSocket.accept();
	    controllerIn = 
		new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));

        } catch (IOException e) {
            System.err.println("Accept failed.");
            System.exit(1);
        }


	try {

	    String fromController = "";
	    String fromServer = "";
	    
	    
	    while (!fromController.equals("bye")) {
		for (int i=0;i<noServers;i++) {
		    if (lcIns[i].ready()) {
			fromServer = lcIns[i].readLine();
			if (fromServer.length()>10)
			    if ((i==0) && (fromServer.substring(0,10).equals("new-txn-id"))) {
				txnId = Long.parseLong(fromServer.substring(11));
				for (int j=1;j<noServers;j++) {
				    lcOuts[j].println("txn " + txnId);
				}
			    }
			System.out.println("Server " + i + ": " + fromServer);
		    }
		}
		
		if (controllerIn.ready()) {
		    fromController = controllerIn.readLine();
		    System.out.println("Controller: " + fromController);
		    StringTokenizer st = new StringTokenizer(fromController);
		    if (st.hasMoreTokens()) {
			String oper = st.nextToken();
			if (st.hasMoreTokens()) {
			    int itemNo = Integer.parseInt(st.nextToken());
			    if (oper.equals("slock")) {
				for (int i=0;i<noSLockServers;i++) {
				    lcOuts[i].println(fromController);
				}
				Lock l = new Lock(txnId, itemNo, Lock.SHARED);
				lockRequests.add(l);
			    } else
			    if (oper.equals("xlock")) {
				for (int i=noServers-1;i>=(noServers-noXLockServers);i--) {
				    lcOuts[i].println(fromController);
				}
				Lock l = new Lock(txnId, itemNo, Lock.EXCLUSIVE);
				lockRequests.add(l);
			    } else
			    if (oper.equals("unlock")) {
				boolean slockFound = false;
				boolean xlockFound = false;
				for (int i=0;i<=lockRequests.size();i++) {
				    Lock l = (Lock) lockRequests.elementAt(i);
				    if ((l.getTxnId()==txnId) && (l.getItemNo()==itemNo) && 
					(l.getLockType()==Lock.SHARED)) {
					lockRequests.remove(i);
					if (!slockFound) {
					    slockFound=true;
					    for (int j=0;j<noSLockServers;j++) {
						lcOuts[j].println(fromController);
					    }
					}
				    }
				    if ((l.getTxnId()==txnId) && (l.getItemNo()==itemNo) && 
					(l.getLockType()==Lock.EXCLUSIVE)) {
					lockRequests.remove(i);
					if (!xlockFound) {
					    xlockFound=true;
					    for (int j=noServers-1;j>=(noServers-noXLockServers);j--) {
						lcOuts[j].println(fromController);
					    }
					}
				    }
				}
			    }
			    else
				for (int k=0;k<noServers;k++)
				    lcOuts[k].println(fromController);

			}
			
			else // only one token
			    if (fromController.equals(LockManagerProtocol.NEW_TXN_COMMAND) ||
				fromController.equals(LockManagerProtocol.COMMIT_COMMAND))
				lcOuts[0].println(fromController);
			    else
				for (int k=0;k<noServers;k++)
				    lcOuts[k].println(fromController);
			
		    }
		}
	    }
	    
	    for (int i=0;i<noServers;i++) {
		lcOuts[i].close();
		lcSockets[i].close();
	    }
	    
	    controllerIn.close();
	    clientSocket.close();
	    serverSocket.close();
	    
	} catch (Exception e) {
	    System.out.println(e);
	}

    }
}
