import java.net.*;
import java.io.*;
import java.util.*;

/**
 * A class to manage the communication protocol with a lock client.
 *
 * @author  Jyrki Nummenmaa
 * @version 0.1
 * @see     LockManager
 * @see     LockManagerProtocol
 */

public class LockManagerProtocol {

    private static final int NO_TXN = -1;
    private static final int NO_ITEM = -1;
    private static final boolean SERVER_CONNECTION = true;
    private static final boolean CLIENT_CONNECTION = false;

    /** Client chared lock command */
    public static final String S_LOCK_COMMAND = "slock";
    /** Client exclusive lock command */
    public static final String X_LOCK_COMMAND = "xlock";
    /** Client unlock command */
    public static final String UNLOCK_COMMAND = "unlock";
    /** Client command for closing connection */
    public static final String BYE_COMMAND = "bye";
    /** Client command for communicating txn id */ 
    public static final String TXN_COMMAND = "txn";
    /** Client command for creating a new txn and requesting the id */ 
    public static final String NEW_TXN_COMMAND = "newtxn";
    /** Client command for initiating deadlock detection */ 
    public static final String SEND_WAITS_FOR_COMMAND = "sendwaitsfor";
    /** Client command for 2PC initiation */ 
    public static final String COMMIT_COMMAND = "commit";
    /** Server command for 2PC vote request */ 
    public static final String TWO_PC_VOTE_REQ_COMMAND = "2PCvotereq";
    /** Server command for 2PC vote yes */ 
    public static final String TWO_PC_YES_COMMAND = "2PCyes";
    /** Server command for 2PC vote no */ 
    public static final String TWO_PC_NO_COMMAND = "2PCno";
    /** Server command for 2PC abort decision */ 
    public static final String TWO_PC_ABORT_COMMAND = "2PCabort";
    /** Server command for 2PC commit decision */ 
    public static final String TWO_PC_COMMIT_COMMAND = "2PCcommit";


    private long txnId = NO_TXN;

    private boolean connectionType = CLIENT_CONNECTION; // Default

    private LockManager lm;

    public LockManagerProtocol(LockManager lm) {
	this.lm = lm;
    }


    /** 
     * Pulls a single lock notification from the notify queue
     * and creates the response string for client.
     */
    public String checkNotify() {
	int itemNo = lm.notifyLock(txnId, Lock.EXCLUSIVE);
	if (itemNo!=LockManager.NO_ITEM)
	    return "xlocked " + itemNo + " - gave queued exclusive lock";
	itemNo = lm.notifyLock(txnId, Lock.SHARED);
	if (itemNo!=LockManager.NO_ITEM)
	    return "slocked " + itemNo + " - gave queued shared lock";
	return null;
    }


    /** 
     * Processes an input string from the client. Calls the respective
     * LockManager's lock management methods and communicates the
     * result to the client.
     */
    public String processInput(String theInput) {

        String theOutput = null;

	if (theInput==null) {
	    return "Connection established";
	}
	
	if (lm.getDebug()>0)
	    if (txnId!=NO_TXN)
		System.out.println("Txn " + txnId + " says: " + theInput);
	    else
		System.out.println("Got input: " + theInput);

	StringTokenizer st = new StringTokenizer(theInput);
	
	String command = "";
	int itemNo = NO_ITEM;

	try {
	    if (st.hasMoreTokens())
		command = st.nextToken();
	    if (command.equals(BYE_COMMAND)) 
		theOutput = BYE_COMMAND;
	    else if (command.equals(TXN_COMMAND)) {
		txnId = Long.parseLong(st.nextToken());
		lm.recordTxn(txnId);
		theOutput = "old-txn-id " + txnId + " confirmed.";
	    }
	    else if (command.equals(NEW_TXN_COMMAND)) {
		txnId = lm.newTxn();
		theOutput = "new-txn-id " + txnId;
	    }
	    else if (command.equals(TWO_PC_VOTE_REQ_COMMAND)) {
		long voteTxnId = Long.parseLong(st.nextToken());
		if (lm.canCommit(voteTxnId))
		    theOutput = TWO_PC_YES_COMMAND;
		else {
		    lm.rollback(txnId);
		    theOutput = TWO_PC_NO_COMMAND;
		}
	    }
	    else if (command.equals(TWO_PC_COMMIT_COMMAND)) {
		long commitTxnId = Long.parseLong(st.nextToken());
		lm.commit(commitTxnId);
		theOutput = "-Committed.";
	    }
	    else if (command.equals(TWO_PC_ABORT_COMMAND)) {
		long rollbackTxnId = Long.parseLong(st.nextToken());
		lm.rollback(rollbackTxnId);
		theOutput = "-Aborted";
	    }
	    else if (command.equals(COMMIT_COMMAND)) {
		new TwoPCCoordThread(lm, txnId).start();
		txnId = NO_TXN;
		theOutput = "-Commit initiated.";
	    }
	    /* else if (command.equals(SEND_WAITS_FOR_COMMAND)) {
		Vector wfi = lm.getWaitsForInfo();
		StringBuffer sb = new StringBuffer();
		sb.append(wfi.size() + "\n");
		for (int i=0;i<wfi.size();i++) {
		    WaitsFor wf = (WaitsFor) wfi.elementAt(i); 
		    sb.append(wf.toString() + "\n");
		}
		System.out.println(sb);
		theOutput = new String(sb);
		}	*/	    
	    else if (command.equals(S_LOCK_COMMAND)) {
		itemNo = Integer.parseInt(st.nextToken());		
		Lock l = new Lock(txnId, itemNo, Lock.SHARED);
		int success = lm.lock(l);
		if (success == LockManager.LOCKED)
		    theOutput = "slocked " + itemNo;
		else
		    if (success == LockManager.QUEUED)
			theOutput = "queued " + itemNo + " for a shared lock.";
		    else // btw, this should never happen:
			theOutput = "denied lock for item " + itemNo;
	    }
	    else if (command.equals(X_LOCK_COMMAND)) {
		itemNo = Integer.parseInt(st.nextToken());
		Lock l = new Lock(txnId, itemNo, Lock.EXCLUSIVE);
		int success = lm.lock(l);
		if (success == LockManager.LOCKED)
		    theOutput = "xlocked " + itemNo;
		else
		    if (success == LockManager.QUEUED)
			theOutput = "queued " + itemNo + " for a shared lock.";
		    else // this should also never happen
			theOutput = "denied lock for item " + itemNo;
	    } else if (command.equals(UNLOCK_COMMAND)) {
		itemNo = Integer.parseInt(st.nextToken());		
		boolean success = lm.unlock(txnId, itemNo);
		if (success)
		    theOutput = "unlocked " + itemNo;
		else
		    theOutput = "unlocking " + itemNo + " failed - you did not have the lock";
	    } else 
		theOutput = "unknown command: " + command;

	}
	catch (Exception e) {
	    theOutput = "error in performing command: " + command;;
	    System.out.println(e);
	}


        return theOutput;
    }
}
